
# (C) 2016 Guillaume Postic
# MAIDEN v1.3

use strict;
#use warnings;

sub all_atoms_or_CA{

    my $ATM_or_CA = shift;
    my $any_atom = shift;
    my @array;
    @array = qw(A C D E F G H I K L
                M N P Q R S T V W Y)
                if ($ATM_or_CA eq 'alpha_C');
    
    @array = qw(A-N A-CA A-CB A-C A-O R-N R-CA R-CB R-CG
                R-CD R-NE R-CZ R-NH1 R-NH2 R-C R-O N-N
                N-CA N-CB N-CG N-OD1 N-ND2 N-C N-O D-N
                D-CA D-CB D-CG D-OD1 D-OD2 D-C D-O C-N
                C-CA C-CB C-SG C-C C-O Q-N Q-CA Q-CB Q-CG
                Q-CD Q-OE1 Q-NE2 Q-C Q-O E-N E-CA E-CB
                E-CG E-CD E-OE1 E-OE2 E-C E-O G-N G-CA
                G-C G-O H-N H-CA H-ND1 H-CG H-CB H-NE2
                H-CD2 H-CE1 H-C H-O I-N I-CA I-CB I-CG2
                I-CG1 I-CD1 I-C I-O L-N L-CA L-CB L-CG
                L-CD1 L-CD2 L-C L-O K-N K-CA K-CB K-CG
                K-CD K-CE K-NZ K-C K-O M-N M-CA M-CB M-CG
                M-SD M-CE M-C M-O F-N F-CA F-CB F-CG
                F-CD1 F-CD2 F-CE1 F-CE2 F-CZ F-C F-O P-N
                P-CA P-CD P-CB P-CG P-C P-O S-N S-CA S-CB
                S-OG S-C S-O T-N T-CA T-CB T-OG1 T-CG2
                T-C T-O W-N W-CA W-CB W-CG W-CD2 W-CD1
                W-NE1 W-CE2 W-CE3 W-CZ2 W-CZ3 W-CH2 W-C W-O
                Y-N Y-CA Y-CB Y-CG Y-CD1 Y-CD2 Y-CE1 Y-CE2
                Y-CZ Y-OH Y-C Y-O V-N V-CA V-CB V-CG1
                V-CG2 V-C V-O)
                if ($ATM_or_CA eq 'all_atm');

    @array = ('x') if ($any_atom);

    return @array;
}


sub create_hash_of_arrays{
# Arguments:
#     None
#
# Returns:
#     an hash: for each key (e.g. I_E_A_G, which means that ALA is inside the membrane and
#     GLY outside), the associated value is a reference to an anonymous array

    my %hash;
    my $ATM_or_CA = shift;
    my $any_location = shift;
    my $any_atom = shift;

    # 1st step: create combinations of 2 amino acids
    # each time, the 2 amino acids are 2 elements of a list
    # and this list is "pushed" in an array called @twoAAs
    my @the20AAs = all_atoms_or_CA($ATM_or_CA, $any_atom);

    my @twoAAs;
    for (my $i = 0; $i <= $#the20AAs; $i++){
        for (my $j = $i; $j <= $#the20AAs; $j++){
            push (@twoAAs, ["$the20AAs[$i]", "$the20AAs[$j]"]);
        }
    }

    my @states_in_membrane = ($any_location) ? ('x_x') : qw(E_E I_I E_I I_E);

    # For the "E_I" couple (we could have chosen "I_E"),
    # there is no push when the 2 AAs are the same
    foreach my $state (@states_in_membrane){
        for (my $j = 0; $j <= $#twoAAs; $j++){
            unless($state eq 'E_I' && $twoAAs[$j][0] eq $twoAAs[$j][1]){
                $hash{"$state\_$twoAAs[$j][0]_$twoAAs[$j][1]"} = [];
            }
        }
    }

    return %hash;
}


sub find_best_model{
# Counts how many models are present in the array containing a PDB file;
# If several, this function finds the best one, its start and end lines;
# If none, the start line index is 0 and the end line is the last element of
# the array (i.e. the entire file will be read)
#
# Arguments:
#     an array containing a PDB file
#
# Returns:
#     two indices, corresponding to the start and the end of the model in
#     the PDB file

    my @pdb_file = @{$_[0]};
	my $models = 0;

	for (my $i = 0; $i <= $#pdb_file; $i++){
		$models++ if ($pdb_file[$i] =~ m/^MODEL\s*\d+/);
	}

	my ($start, $end) = (0) x 2;

	if ($models > 0){
# STDERRRRRR
                print "Input coordinate file contains several models\n";
                print "For a multimeric protein, please make sure that each monomer is represented by a chain, not by a model.\n";

		my $bestModel;

		for (my $i = 0; $i <= $#pdb_file; $i++){
			if ($pdb_file[$i] =~ m/BEST\sREPRESENTATIVE\sCONFORMER\sIN\sTHIS\s
			                     ENSEMBLE\s:\s([0-9a-zA-Z]+)/x){

				$bestModel = ($1 eq 'NULL') ? 1 : $1;

				my $j = $i;
				while ($pdb_file[$j] !~ m/^MODEL\s*$bestModel/){
					$j++;
				}

				$start = $j+1;
				my $k = $start;

				while ($pdb_file[$k] !~ m/^ENDMDL/){
					$k++;
				}

				$end = $k-1; # not -2 in the case of there is no TER
			}
		}

		# several models, but if none selected:
		if (!$end){
                        no warnings;

			$bestModel = 1; # arbitrary

			my $i = 0;
			while ($pdb_file[$i] !~ m/^MODEL\s*$bestModel\s+/){
				$i++;
			}

			$start = $i+1;
			my $j = $start;

			while ($pdb_file[$j] !~ m/^ENDMDL/ && $j < 50000){
				$j++;
			}

			$end = $j-1; # not -2 in the case of there is no TER

                        $end = $#pdb_file if ($j == 50000);
		}
	}

    # if no model:
    else{
        my $i = 0;
        my $record = substr($pdb_file[$i],0,6);
        while ($record !~ m/ATOM|HETATM/){
            $i++;
            $record = substr($pdb_file[$i],0,6);
	    print "Error: input probably not a proper PDB file\n" if ($i > 50000);
	    die "Error: input probably not a proper PDB file\n" if ($i > 50000);
        }
        $start = $i;
	$end = $#pdb_file;
    }

    return ($start, $end);
}


sub intern_residues{
# For a chain, creates a list of all residues that are in the membrane,
# based on the PDBTM XML file. It also creates a list of the external residues.
#
# Arguments:
#    the array returned by open_xml_file()
#    a protein chain
#
# Returns:
#    two arrays (refs) that contain intra- and extra-membrane residues numbers

    my $chain = $_[0];
    my @xmlFile = @{$_[1]};
    my %hash = %{$_[2]} if ($_[2]);

    my @residues_I = ();
    my @residues_E = ();

    if (!$_[2]){
        # Delimiting the XML file (array)
        my ($begXML, $endXML);
    
        for (my $i = 0; $i <= $#xmlFile; $i++){
            $begXML = $i if ($xmlFile[$i] =~ m/<CHAIN CHAINID="$chain"/);
        }
    
        my $j = $begXML;
    
        while ($xmlFile[$j] !~ m/CHAIN>/){
            $j++;
        }
    
        $endXML = $j++;
    
        # Read the XML file (array) within the delimitations
        for (my $k = $begXML; $k <= $endXML; $k++){
            if ($xmlFile[$k] =~ m/<REGION\s+seq_beg="[0-9]+"\s+
                                  pdb_beg="([0-9]+)"\s+seq_end="[0-9]+"\s+
                                  pdb_end="([0-9]+)"\s+type="([12BHCILU])"/x){
    
                my $pdb_beg_var = $1;
                my $pdb_end_var = $2;
                my $res_state = $3;
    
                if ($res_state ~~ [qw(H B C I)]){
                    my @array = map {$_} ($pdb_beg_var..$pdb_end_var);
                    push (@residues_I, @array);
                }
                elsif ($res_state !~ [qw(U L)]){ # L is intra- but not trans-membrane
                    my @array = map {$_} ($pdb_beg_var..$pdb_end_var);
                    push (@residues_E, @array);
                }
            }
        }
    }

    else{
 
        foreach (@{$hash{$chain}}){
            my ($beg, $end) = split('-', $_);
            my @array = map {$_} ($beg..$end);
            push (@residues_I, @array);
        }

        # @residues_E remains empty
    }

    return (\@residues_I, \@residues_E);
}


sub chains_check_pdb{
# Finds chains that are present in the PDB file
#
# Arguments:
#     the start and end indices returned by find_best_model
#     the array (reference) returned by open_pdb_file
#
# Returns:
#     an array with the chains in the PDB file

    my $CG = shift; # 0 if CA; 1 if coarse-grained
    my $start = shift;
    my $end = shift;
    my @pdb_file = @{$_[0]};

    my $atomType = $CG ? 'BB' : 'CA';

    my $regex = qr/^ATOM\s*-?\d+\s*$atomType\s*[A-Za-z0-9]?[A-Z]{3}
                   \s([a-zA-Z0-9_ ]?)\s*-?\d{1,4}\s*-?\d+\.\d+
                   \s*-?\d+\.\d+\s*-?\d+\.\d+/x;


    my @chainListPDB;

    for (my $i = $start; $i <= $end; $i++){
        chomp $pdb_file[$i];
        if ($pdb_file[$i] =~ /$regex/){
		push (@chainListPDB, $1) unless (grep(/^$1$/, @chainListPDB));
	}
    }

    return @chainListPDB;
}


sub chains_check_xml{
# Finds chains that are present in the XML file
#
# Arguments:
#     the array (reference) returned by open_xml_file
#
# Returns:
#     an array with the chains in the XML file

    my @chainListXML = ();

    foreach my $line (@{$_[0]}){
        push (@chainListXML, $1) if ($line =~ m/<CHAIN\sCHAINID="([A-Za-z0-9_]*)"/);
    }

    return @chainListXML;
}


sub compare_chains{
# Compares the arrays returned by chains_check_pdb() and chains_check_xml()
#
# Arguments:
#     the arrays (references) returned by chains_check_pdb() and chains_check_xml()
#     the processed PDB file
#
# Returns:
#     an array of all the chains that are shared by the PDB and XML files
#     stops the program when a chain is found in the XML and not in the PDB

	my @chainListPDB = @{$_[0]};
	my @chainListXML = @{$_[1]};

	my $entpdb = $_[2];
	my @chainList;

	for (my $i = 0; $i <= $#chainListXML; $i++){
		my $count = 0;

		for (my $j = 0; $j <= $#chainListPDB; $j++){
			if ($chainListXML[$i] eq $chainListPDB[$j]){
				push (@chainList, $chainListXML[$i]);
				$count++;
			}
		}

		print "\nError: chain name $chainListXML[$i] not found in $entpdb\n" if (!$count);
		die "\nError: chain name $chainListXML[$i] not found in $entpdb\n" if (!$count);
	}

	return @chainList;
}


sub chain_first_residue{
# For a given chain of a PDB file, finds the first residue number
#
# Arguments:
#     the "start index" returned by find_best_model
#     the PDB file in an array (returned by open_pdb_file)
#     the chain name
#
# Returns:
#     the first residue number

    my $CG = shift; # 0 if CA; 1 if coarse-grained
    my $chain = $_[2];
    my @pdb_file = @{$_[1]};
    my $line_num = $_[0]; # initialized by $start
    my $first_num = 0;

    my $atomType = $CG ? 'BB' : 'CA';

    my $regex =  qr/^(ATOM)\s*(-?\d+)\s*$atomType\s*[A-Za-z0-9]?
                    ([A-Z]{3})\s([$chain])
                    \s*(-?\d{1,4})\s*(-?\d+\.\d+)\s*
                    (-?\d+\.\d+)\s*(-?\d+\.\d+)/x;

    while ($pdb_file[$line_num] !~ /$regex/){
        $line_num++;
    }

    $first_num = $5 if ($pdb_file[$line_num] =~ /$regex/);
    return $first_num;
}


sub get_coordinates{
# For the specified chain, reads the PDB array (the good model, if any);
# For each alpha carbon,
# >>> it gives a new ("standard") so the different chains of the PDB file are
# numbered in continue (and the first residue number will be 1)
# >>> it extracts the residue name, residue number, x y z coordinates;
# These values are gathered in an array;
# In a hash (parameter), the standard number will be the key associated to the
# reference of this array
#
# Arguments:
#     an empty hash
#     an array returned by open_pdb_file()
#     a chain name
#     an array returned by intern_residues()
#     the $start and $end returned by find_best_model()
#     first residue number either returned by chain_first_residue() or by this very function
#     the standard number of the last residue
#     the PDB entry
#
# Returns:
#     the standard number of the last residue

    my $hash_ref = shift;
    my @pdb_file = @{$_[0]};
    my $query_chain = $_[1];
    my @residues_I = @{$_[2]};
    my ($start, $end, $first_num, $last_std_num,
        $pdbCode, $ATM_or_CA, $any_location, $any_atom, $verbose) = splice(@_, 3, 9);

    my $std_num;
    my %hash_3to1 = create_hash_3to1();

    for (my $i = $start; $i <= $end; $i++){
        my $state = ($any_location) ? "x" : "E";
        my $previous_num = $last_std_num;

        # Extracting info from a PDB file line
        my ($record, $atom_num, $atom_name, $res_name, $chain, $res_num, $x, $y, $z);
	eval{
        	($record, $atom_num, $atom_name, $res_name, $chain, $res_num, $x, $y, $z) = 
		map {s/ //g; $_} (unpack('@0A6 @6A5 @12A4 @17A3 @21A1 @22A4 @30A8 @38A8 @46A8', $pdb_file[$i]));
	};

	if ($@){
		unless ($pdb_file[$i] =~ m/^END|^TER/){
			print "\nError: Line not complying with PDB coordinate file standards:\n","$pdb_file[$i]\n";
			die "\nError! Line not complying with PDB coordinate file standards:\n","$pdb_file[$i]\n";
		}
	}

        $atom_name = 1 if (!$atom_name); # this can happen, e.g. when last line is only "TER"

        $atom_name =~ s/OXT/O/g;
	$res_name = $hash_3to1{$res_name} if ($res_name);

        $chain = ' ' if (!$chain);

        if ($atom_name eq 'CA' or $atom_name eq 'BB' or $ATM_or_CA eq 'all_atm'){
            if ($record eq 'ATOM' and $res_name ne 'X' and $chain eq $query_chain){
                $std_num = $res_num - $first_num + 1 + $last_std_num;

                # chimeric structures (e.g. with lysozyme) must have been
                # processed or it will lead to a warning
                #     checking whether a residue with this standard number (of
                #     this chain) has already been added in the hash or not;
                #     if it has, then it is a problem and this residues won't
                #     be processed a second time
                #         checking whether the residue is inside the membrane
                if ($std_num > $previous_num){
                    unless ($any_location){
                        for (my $j = 0; $j <= $#residues_I; $j++){
                            $state = 'I' if ($residues_I[$j] == $res_num);
                        }
                    }

                    if ($state eq 'I'){
                        my $key = $std_num;
                        if ($ATM_or_CA eq 'all_atm'){
                            $res_name = "$res_name-$atom_name";
                            my $std_num_10E5 = $std_num * 10E5; 
                            $key = $std_num_10E5 + $atom_num;
                        }
                       
                        $res_name = 'x' if ($any_atom);
                        ${$hash_ref}{$key} = [$res_name, $x, $y, $z, $state, $pdbCode, $key, $res_num, $chain]; #v1.4: +$res_num & $chain
                    }
                }

                else{
                    $res_name = "$res_name-$atom_name" if ($ATM_or_CA eq 'all_atm');
                    print STDERR "Warning: problem in the sorting of the residues numbers\n",
                          "\tnumber $std_num (name: $res_name coord.: $x $y $z) follows $previous_num\n";
                }

                $previous_num = $std_num;
            }
        }
    }

    my $ok_or_not = ((keys(%$hash_ref)) > 0) ? "OK\n" : "no residues\n";
    print "Chain: $query_chain ... $ok_or_not" if ($verbose);

    # if the interaction between the last residue of one chain and the
    # first residue of the next chain have to be considered, the they will
    # have to be separated by 4, since we do not consider interactions between
    # residues i and i+3 or i+2 or i+1
    $last_std_num = $std_num + 4;

    return $last_std_num;
}


sub create_hash_3to1{

    my %hash_3to1;
    my @array = qw(ALA A CYS C ASP D GLU E PHE F GLY G HIS H ILE I LYS K LEU L
                   MET M ASN N PRO P GLN Q ARG R SER S THR T VAL V TRP W TYR Y
                   UNK X);
    for (my $i = 0; $i < $#array; $i+= 2){
        my $j = $i + 1;
        $hash_3to1{$array[$i]} = $array[$j];
    }

    return %hash_3to1;
}


sub read_file{

    open (my $fh, '<', $_[0])
        or die "\nError in $0 while reading $_[0]:\n\t$!\n";
    <$fh>;
}


sub residue_position{

    my $AA = shift;
    my $array_ref = shift;

    my $naa = 0;

    while ($AA ne ${$array_ref}[$naa]){
        $naa++;
    }	

    return $naa;
}


sub extract_input{
    my %hash;
    
    my @segments = split ('_', $_[0]);
    
    for (@segments){
        my $seg = $_;
        my $chain = substr($seg,0,1);
        $seg =~ s/^.//;
    
        if (exists $hash{$chain}){
            push(@{$hash{$chain}}, $seg);
        }
        else{
            $hash{$chain} = [];
            push(@{$hash{$chain}}, $seg);
        }
    }

    return %hash;
}

1;
__END__

